<?php
// $Id: stats_plot_user_activity.inc,v 1.5 2004/08/11 05:53:15 njivy Exp $
/**
 * Overview
 * --------
 * This file is included by stats_plot.module.
 *
 * This file generates six interrelated plots and charts.  In this arrangement, charts
 *  that have one axes in common are placed adjacent to each other.  For example, if two 
 * charts have identical horizontal axis in time, then they are positioned above and below 
 * each other.  By doing this, data in both plots can easily be found for a single point in 
 * time.  To reinforce the idea that the two adjacent plots are related, a marginal 
 * distribution is plotted between them: one vertical line per data point.  The marginal has 
 * the added benefit of depicting the density of data points in one dimension as well as 
 * reinforce the relatedness of adjacent plots.  The plots and marginal distributions look 
 * something like a rug, so this is called a rug plot.
 * 
 * Example 1:
 * ______________
 * |        .   |
 * | .       .  |  <-- scatter plot of K vs X
 * |       .    |
 * --------------
 *   |  || |||     <-- marginal distribution of X
 * ______________
 * |            |
 * | .          |
 * |     .      |  <-- scatter plot of J vs X
 * |    .       |
 * |            |
 * --------------
 * 
 * In particular, this file generates six plots that relate to user activity on a typical
 * Drupal website.  This file assumes that multiple people participate in submitting content
 * and comments, that these people are registered users, and that basic statistical logging is
 * enabled in Drupal.
 * 
 * The six plots are these:
 * 1) Number of hours between successive posts in a thread vs. number of posts in the thread
 * 2) Number of characters in a post vs. number of posts in a thread
 * 3) Number of participating users vs. number of posts in a thread
 * 4) Number of page requests vs. time
 * 5) Number of characters in a post vs. time
 * 6) Number of users online vs. time
 * 
 * As you can see, each plot has at least one axes in common with another plot.  The rug plot
 * will look something like Example 2.
 * 
 * Example 2:
 * _____________   __________
 * |           |   |        |
 * |    #1     |   |  #4    |
 * |           |   |        |
 * -------------   ----------
 *    | |  ||       | |||  | 
 * _____________   __________
 * |           | - |        |
 * |           | - |        |
 * |    #2     |   |  #5    |
 * |           | - |        |
 * |           |   |        |
 * -------------   ----------
 *    | |  ||       | |||  | 
 * _____________   __________
 * |           |   |        |
 * |    #3     |   |  #6    |
 * |           |   |        |
 * |           |   |        |
 * -------------   ----------
 * 
 * All of the plots in the left column have the same horizontal axes.  Likewise for the right
 * column.  The middle row has the same vertical axes.
 * 
 * The tricky part of this file is two-fold:
 * 1) Gathering the data from Drupal: The flexinode.module can complicate this.
 * 2) Efficiently plotting the data: The algorithm and data structures must be carefully
 *    tuned to minimize overhead and computational complexity.  For the sake of reducing
 *    computational effort, plotted data will only come from the recent past.  This can
 *    be accomplished by either removing old log entries or forcing this file to use
 *    only the most recent data.
 * 
 * The GD image libraries of PHP will likely be used. See [1] and [2] for more information.
 * Inspiration for this work comes from [3].
 * 
 * Helpful comments and insight are desired.
 * 
 * 
 * Author
 * ------
 * Nic Ivy (http://njivy.org)
 * 
 * 
 * References
 * ----------
 * [1] http://www.php.net/manual/en/ref.image.php
 * [2] http://builder.com.com/5100-6371-5092227.html
 * [3] Tufte, Edward R. "The Visual Display of Quantitative Information". 1983. 
 *     Graphics Press: Cheshire, Connecticut. p 135.
 *
 */

/**
 * This function is included by stats_plot.module's hook_settings() implementation.
 */
function stats_plot_user_activity_settings() {
  $variables = array('%run_time', '%site_name');
  $point_sizes = array(1=>'1 px', 2=>'2px', 3=>'3px', 4=>'4px', 5=>'5px');

  $tmp_output .= form_textfield(t('Title'), 'stats_plot_user_activity_title', variable_get('stats_plot_user_activity_title', ''), 64, 128, t('Available variables include: %a', array('%a'=>implode(', ', $variables))));
  $tmp_output .= form_textfield(t('Footer'), 'stats_plot_user_activity_footer', variable_get('stats_plot_user_activity_footer', ''), 64, 128, t('Available variables include: %a', array('%a'=>implode(', ', $variables))));
  $tmp_output .= form_textfield(t('Full path to output file'), 'stats_plot_user_activity_output_file', variable_get('stats_plot_user_activity_output_file', ''), 64, 128, t('Please be sure the file is writeable.'));
  $tmp_output .= form_textfield(t('Data color'), 'stats_plot_user_activity_data_color', 
                                  variable_get('stats_plot_user_activity_data_color', '#000000'), 
                                  7, 7, t('Please provide a valid hexidecimal HTML color value.'));
  $tmp_output .= form_textfield(t('Background color'), 'stats_plot_user_activity_background_color', 
                                  variable_get('stats_plot_user_activity_background_color', '#FFFFFF'), 
                                7, 7, t('Please provide a valid hexidecimal HTML color value.'));
  $tmp_output .= form_textfield(t('Plot background color'), 'stats_plot_user_activity_plot_background_color', 
                                  variable_get('stats_plot_user_activity_plot_background_color', '#EFEFEF'), 
                                7, 7, t('Please provide a valid hexidecimal HTML color value.'));
  $tmp_output .= form_textfield(t('Border color'), 'stats_plot_user_activity_border_color', 
                                  variable_get('stats_plot_user_activity_border_color', '#000000'), 
                                7, 7, t('Please provide a valid hexidecimal HTML color value.'));
  $tmp_output .= form_textfield(t('Margin color'), 'stats_plot_user_activity_margin_color', 
                                  variable_get('stats_plot_user_activity_margin_color', '#000000'), 
                                7, 7, t('Please provide a valid hexidecimal HTML color value.'));
  $tmp_output .= form_textfield(t('Text color'), 'stats_plot_user_activity_text_color', 
                                  variable_get('stats_plot_user_activity_text_color', '#000000'), 
                                7, 7, t('Please provide a valid hexidecimal HTML color value.'));
  $tmp_output .= form_select(t('Point size'), 'stats_plot_user_activity_point_size', variable_get('stats_plot_user_activity_point_size', 2), $point_sizes, t('Specify the size of each data point.'));
  $tmp_output .= form_textfield(t('Padding'), 'stats_plot_user_activity_padding', variable_get('stats_plot_user_activity_padding', 30), 3, 3, t('Specify the padding, in pixels, around the data area and inside each plot\'s borders.  The tick marks and axes labels go in this area.'));
  $tmp_output .= form_textfield(t('Left column width'), 'stats_plot_user_activity_col_1', variable_get('stats_plot_user_activity_col_1', 200), 3, 4, t('Specify the width, in pixels, of the left-hand column.'));
  $tmp_output .= form_textfield(t('Right column width'), 'stats_plot_user_activity_col_2', variable_get('stats_plot_user_activity_col_2', 350), 3, 4, t('Specify the width, in pixels, of the right-hand column.'));
  $tmp_output .= form_textfield(t('Top row height'), 'stats_plot_user_activity_row_1', variable_get('stats_plot_user_activity_col_1', 300), 3, 4, t('Specify the height, in pixels, of the top row.'));
  $tmp_output .= form_textfield(t('Middle row height'), 'stats_plot_user_activity_row_2', variable_get('stats_plot_user_activity_col_1', 250), 3, 4, t('Specify the height, in pixels, of the middle row.'));
  $tmp_output .= form_textfield(t('Bottom row height'), 'stats_plot_user_activity_row_3', variable_get('stats_plot_user_activity_col_1', 200), 3, 4, t('Specify the height, in pixels, of the bottom row.'));

  $output = form_group(t('User activity plot settings'), $tmp_output);

  return $output;
}

/**
 * This function is called when the plot is updated.  It contains the configuration
 * information for the individual plots and assembles all the pieces.
 */
function stats_plot_user_activity() {
  $now = time();
  
  ////
  //// Start the timer. This is for debugging, of course.
  ////
  //$timer_1 = explode(' ', microtime());
  
  /////////////////////////////////////// GENERAL CONFIGURATION ////////////////////////////////////////////////////
  //
  // Plots:
  // 1) Time between posts vs. number of posts in thread
  // 2) Number of characters in post vs. number of posts in thread
  // 3) Number of participating users vs. number of posts in thread
  // 4) Node requests vs. hour of day
  // 5) Number of characters in post vs. hour of day
  // 6) Number of users online vs. hour of day
  //
  $font_file = variable_get('stats_plot_font_file', '');
  
  $output_file = variable_get('stats_plot_user_activity_output_file', ''); // The type of file must be PNG.
  
  $time_window = variable_get('stats_plot_days_of_data', 30) * 24 * 60 * 60; // seconds equivalent to 30 days
  $rug_plot_title = strtr(variable_get('stats_plot_user_activity_title', ''), array('%run_time'=>format_date($now, 'small'), '%site_name'=>variable_get('site_name', 'drupal')));
  $rug_plot_footer = strtr(variable_get('stats_plot_user_activity_footer', ''), array('%run_time'=>$now, '%site_name'=>variable_get('site_name', 'drupal')));
  
  $attributes = array('padding'=>variable_get('stats_plot_user_activity_padding', 30),
                      'point width'=>variable_get('stats_plot_user_activity_point_size', 2),
                      'point height'=>variable_get('stats_plot_user_activity_point_size', 2),
                      'data color'=>variable_get('stats_plot_user_activity_data_color', '#000000'),
                      'background color'=>variable_get('stats_plot_user_activity_background_color', '#FFFFFF'),
                      'border color'=>variable_get('stats_plot_user_activity_border_color', '#000000'),
                      'margin color'=>variable_get('stats_plot_user_activity_margin_color', '#000000'),
                      'plot background color'=>variable_get('stats_plot_user_activity_plot_background_color', '#EFEFEF'),
                      'text color'=>variable_get('stats_plot_user_activity_text_color', '#000000'));
  
  $plot_sizes['column 1'] = variable_get('stats_plot_user_activity_col_1', 200);
  $plot_sizes['column 2'] = variable_get('stats_plot_user_activity_col_2', 300);
  $plot_sizes['row 1'] = variable_get('stats_plot_user_activity_row_1', 300);
  $plot_sizes['row 2'] = variable_get('stats_plot_user_activity_row_2', 250);
  $plot_sizes['row 3'] = variable_get('stats_plot_user_activity_row_3', 200);
  
  $plot_ranges[1]['x min'] = 1; // number of posts
  $plot_ranges[1]['x max'] = 50;
  $plot_ranges[1]['y min'] = 60; // seconds between posts
  $plot_ranges[1]['y max'] = 8640 * 2 * 3600;
  $plot_ranges[1]['y log'] = 1; // logarithmic scale
  $plot_labels[1]['title'] = 'Time between posts vs. length of thread';
  $plot_labels[1]['x marks'] = array(1=>'1 post', 5=>'', 10=>'', 15=>'', 20=>'', 25=>25, 30=>'', 35=>'', 40=>'', 45=>'', 50=>50);
  $plot_labels[1]['y marks'] = array(60=>'1 min', 1800=>'30 min', 3600=>'1 hr', 12*3600=>'12 hr',
                                     24*3600=>'1 dy',
                                     168*3600=>'1 wk', 
                                     720*3600=>'1 mo', 720*3600*6=>'6 mo',
                                     8640*3600=>'1 yr');
  $plot_margins[1] = array('left'=>0, 'right'=>0, 'top'=>0, 'bottom'=>1);
  
  $plot_ranges[2]['x min'] = $plot_ranges[1]['x min']; // number of posts
  $plot_ranges[2]['x max'] = $plot_ranges[1]['x max'];
  $plot_ranges[2]['y min'] = 1; // characters in post
  $plot_ranges[2]['y max'] = 10000;
  $plot_ranges[2]['y log'] = 1;
  $plot_labels[2]['title'] = 'Length of post vs. length of thread';
  $plot_labels[2]['x marks'] = $plot_labels[1]['x marks'];
  $plot_labels[2]['y marks'] = array(1=>'1 char', 5=>5, 10=>10,
                                     50=>'', 100=>100,
                                     500=>'', 1000=>1000,
                                     5000=>'', 10000=>'10000');
  $plot_margins[2] = array('left'=>0, 'right'=>1, 'top'=>0, 'bottom'=>1);
  
  $plot_ranges[3]['x min'] = $plot_ranges[1]['x min']; // number of posts
  $plot_ranges[3]['x max'] = $plot_ranges[1]['x max'];
  $plot_ranges[3]['y min'] = 1; // number of users
  $plot_ranges[3]['y max'] = 10;
  $plot_labels[3]['title'] = 'Participating users vs. length of thread';
  $plot_labels[3]['x marks'] = $plot_labels[1]['x marks'];
  $plot_labels[3]['y marks'] = array(1=>'1 user', 2=>'', 3=>'', 4=>'', 5=>5, 6=>'', 7=>'', 8=>'', 9=>'', 10=>10);
  $plot_margins[3] = array('left'=>0, 'right'=>0, 'top'=>0, 'bottom'=>0);
  
  $plot_ranges[4]['x min'] = -720; // hours from now
  $plot_ranges[4]['x max'] = 0;
  $plot_ranges[4]['y min'] = 1; // page requests
  $plot_ranges[4]['y max'] = 500;
  $plot_ranges[4]['y log'] = 1;
  $plot_labels[4]['title'] = 'Page requests per hour';
  $plot_labels[4]['x marks'] = array((-24*30)=>'30 days ago', (-24*21)=>'3 weeks ago', (-24*14)=>'2 weeks ago', (-24*7)=>'1 week ago', 0=>0);
  $plot_labels[4]['y marks'] = array(1=>'1 request', 5=>'', 
                                     10=>10, 50=>'', 
                                     100=>100, 500=>500);
  $plot_margins[4] = array('left'=>0, 'right'=>0, 'top'=>0, 'bottom'=>0);
  
  $plot_ranges[5]['x min'] = $plot_ranges[4]['x min'];
  $plot_ranges[5]['x max'] = $plot_ranges[4]['x max'];
  $plot_ranges[5]['y min'] = $plot_ranges[2]['y min'];
  $plot_ranges[5]['y max'] = $plot_ranges[2]['y max'];
  $plot_ranges[5]['y log'] = 1;
  $plot_labels[5]['title'] = 'Length of posts per hour';
  $plot_labels[5]['x marks'] = $plot_labels[4]['x marks'];
  $plot_labels[5]['y marks'] = $plot_labels[2]['y marks'];
  $plot_margins[5] = array('left'=>0, 'right'=>0, 'top'=>1, 'bottom'=>1);
  
  $plot_ranges[6]['x min'] = $plot_ranges[4]['x min'];
  $plot_ranges[6]['x max'] = $plot_ranges[4]['x max'];
  $plot_ranges[6]['y min'] = 1;
  $plot_ranges[6]['y max'] = 10;
  $plot_labels[6]['title'] = 'Online users per hour';
  $plot_labels[6]['x marks'] = $plot_labels[4]['x marks'];
  $plot_labels[6]['y marks'] = array(0=>'0 users', 1=>'1 user', 2=>'', 3=>'', 4=>'', 
                                     5=>5, 6=>'', 7=>'', 8=>'', 9=>'', 10=>10);
  $plot_margins[6] = array('left'=>0, 'right'=>0, 'top'=>0, 'bottom'=>0);
  
  //
  /////////////////////////////////////// END GENERAL CONFIGURATION ////////////////////////////////////////////////
  
  //
  // NOTICE:
  // If the image returns blank, it is likely that PHP threw an error, warning, or notice
  // of some sort.  This file will not work if E_NOTICE is enabled.
  //
  
  //
  // Select node information and begin building $threads
  //
  $result = db_query('SELECT nid, uid, created, body 
                         FROM node 
                         WHERE status=1 
                           AND '.$now.' - created < '.$time_window.'
                         ORDER BY nid');
  while ($t = db_fetch_object($result)) {
    $threads[$t->nid]['last post'] = $t->created;
    $threads[$t->nid]['timestamp'][0] = $t->created;
    $threads[$t->nid]['active users'][$t->uid] = $t->created;
    $threads[$t->nid]['lengths'][0] = strlen($t->body);
  }
  
  //
  // Select comment information and build $threads
  //
  $result = db_query('SELECT cid, nid, uid, comment, timestamp 
                         FROM comments 
                         WHERE status=0 
                           AND '.$now.' - timestamp < '.$time_window.'
                         ORDER BY cid');
  while ($t = db_fetch_object($result)) {
    $threads[$t->nid]['comments'][$t->cid] = ($t->timestamp - $threads[$t->nid]['last post']); // time between posts
    $threads[$t->nid]['timestamp'][$t->cid] = $t->timestamp; // time of posting
    $threads[$t->nid]['active users'][$t->uid] = $t->timestamp;
    $threads[$t->nid]['lengths'][$t->cid] = strlen($t->comment);
    $threads[$t->nid]['last post'] = $t->timestamp;
  }
  
  //
  // Select access log information and build $requests and $visitors
  //
  $result = db_query('SELECT timestamp, uid FROM accesslog WHERE '.$now.' - timestamp < '.$time_window);
  while ($t = db_fetch_object($result)) {
    $requests[$t->timestamp]++;
    $hours_ago = floor(($t->timestamp - $now) / 3600);
    $visitors[$hours_ago][$t->uid] = 1;
  }
  
  //
  // Generate plot 1: Time between posts vs. number of posts in thread
  //
  if (count($threads) > 0) {
    foreach ($threads as $nid=>$data) {
      $post_count = count($data['comments']) + 1;
      if ($post_count > 1) {
        foreach ($data['comments'] as $cid=>$value) {
          $values[$post_count][$value]++;
        }
      }
    }
  }
  
  $attributes['title'] = $plot_labels[1]['title'];
  $attributes['plot width'] = $plot_sizes['column 1'];
  $attributes['plot height'] = $plot_sizes['row 1'];
  $attributes['x log'] = $plot_ranges[1]['x log'];
  $attributes['y log'] = $plot_ranges[1]['y log'];
  //$attributes['data transparency'] = 3*32;
  
  $plots[1] = stats_plot_render_scatterplot($values, $plot_ranges[1], array('x'=>$plot_labels[1]['x marks'], 'y'=>$plot_labels[1]['y marks']), 
                                         $plot_margins[1], $attributes);
  
  //
  // Generate plot 2: Number of characters in post vs. number of posts in thread
  //
  unset($values);
  if (count($threads) > 0) {
    foreach ($threads as $nid=>$data) {
      $post_count = count($data['lengths']);
      foreach ($data['lengths'] as $cid=>$value) {
        $values[$post_count][$value]++;
      }
    }
  }
  
  $attributes['title'] = $plot_labels[2]['title'];
  $attributes['plot width'] = $plot_sizes['column 1'];
  $attributes['plot height'] = $plot_sizes['row 2'];
  $attributes['x log'] = $plot_ranges[2]['x log'];
  $attributes['y log'] = $plot_ranges[2]['y log'];
  //$attributes['data transparency'] = 3*32;
  
  $plots[2] = stats_plot_render_scatterplot($values, $plot_ranges[2], array('x'=>$plot_labels[2]['x marks'], 'y'=>$plot_labels[2]['y marks']), 
                                         $plot_margins[2], $attributes);
  
  //
  // Generate plot 3: Number of participating users vs. number of posts in thread
  //
  unset($values);
  if (count($threads) > 0) {
    foreach ($threads as $nid=>$data) {
      $post_count = count($data['lengths']);
      $values[$post_count][count($data['active users'])]++;
    }
  }
  
  $attributes['title'] = $plot_labels[3]['title'];
  $attributes['plot width'] = $plot_sizes['column 1'];
  $attributes['plot height'] = $plot_sizes['row 3'];
  $attributes['x log'] = $plot_ranges[3]['x log'];
  $attributes['y log'] = $plot_ranges[3]['y log'];
  //$attributes['data transparency'] = 3*32;
  
  $plots[3] = stats_plot_render_scatterplot($values, $plot_ranges[3], array('x'=>$plot_labels[3]['x marks'], 'y'=>$plot_labels[3]['y marks']), 
                                         $plot_margins[3], $attributes);
  
  //
  // Generate plot 4: Node requests vs. hours before now
  //
  unset($values);
  if (count($requests) > 0) {
    foreach ($requests as $timestamp=>$count) {
      $hours_ago = floor(($timestamp - $now) / 3600);
      $tmp[$hours_ago] += $count;
    }
    foreach($tmp as $key=>$value) {
      $values[$key][$value]++;
    }
    unset($tmp);
  }
  
  $attributes['title'] = $plot_labels[4]['title'];
  $attributes['plot width'] = $plot_sizes['column 2'];
  $attributes['plot height'] = $plot_sizes['row 1'];
  $attributes['x log'] = $plot_ranges[4]['x log'];
  $attributes['y log'] = $plot_ranges[4]['y log'];
  //$attributes['data transparency'] = 3*32; // 0 opaque, 127 transparent
  
  $plots[4] = stats_plot_render_scatterplot($values, $plot_ranges[4], array('x'=>$plot_labels[4]['x marks'], 'y'=>$plot_labels[4]['y marks']), 
                                         $plot_margins[4], $attributes);
  
  //
  // Generate plot 5: Number of characters in post vs. hour of day
  //
  unset($values);
  if (count($threads) > 0) {
    foreach ($threads as $nid=>$data) {
      foreach ($data['lengths'] as $cid=>$strlen) {
        $hours_ago = floor(($data['timestamp'][$cid] - $now) / 3600);
        $values[$hours_ago][$strlen]++;
      }
    }
  }
//  print '<pre>'; print_r($values); print '</pre>';
//  exit;
  
  $attributes['title'] = $plot_labels[5]['title'];
  $attributes['plot width'] = $plot_sizes['column 2'];
  $attributes['plot height'] = $plot_sizes['row 2'];
  $attributes['x log'] = $plot_ranges[5]['x log'];
  $attributes['y log'] = $plot_ranges[5]['y log'];
  //$attributes['data transparency'] = 3*32;
  
  $plots[5] = stats_plot_render_scatterplot($values, $plot_ranges[5], array('x'=>$plot_labels[5]['x marks'], 'y'=>$plot_labels[5]['y marks']), 
                                         $plot_margins[5], $attributes);
  
  //
  // Generate plot 6: Number of users online vs. hour of day
  //
  unset($values);
  if (count($visitors) > 0) {
    foreach ($visitors as $hours_ago=>$uids) {
      $values[$hours_ago][count($uids)]++;
    }
  }
  
  $attributes['title'] = $plot_labels[6]['title'];
  $attributes['plot width'] = $plot_sizes['column 2'];
  $attributes['plot height'] = $plot_sizes['row 3'];
  $attributes['x log'] = $plot_ranges[6]['x log'];
  $attributes['y log'] = $plot_ranges[6]['y log'];
  //$attributes['data transparency'] = 3*32;
  
  $plots[6] = stats_plot_render_scatterplot($values, $plot_ranges[6], array('x'=>$plot_labels[6]['x marks'], 'y'=>$plot_labels[6]['y marks']), 
                                         $plot_margins[6], $attributes);
  
  //
  // Prepare to make the final rug-plot
  //
  $background_color = stats_plot_convert_color($attributes['background color']);
  $text_color = stats_plot_convert_color($attributes['text color']);
  $margin_color = stats_plot_convert_color($attributes['margin color']);
  
  $bounding_box = imageftbbox(10, 0, $font_file, $rug_plot_title, array());
  $width[1] = abs($bounding_box[0]) + abs($bounding_box[2]);
  $height[1] = abs($bounding_box[1]) + abs($bounding_box[5]) + 5;

  $bounding_box = imageftbbox(8, 0, $font_file, $rug_plot_footer, array());
  $width[2] = abs($bounding_box[0]) + abs($bounding_box[2]);
  $height[2] = abs($bounding_box[1]) + abs($bounding_box[5]) + 5;

  //
  // Create the final rug-plot
  //
  $image->width = imagesx($plots[2]) + imagesx($plots[4]);
  $image->height = imagesy($plots[1]) + imagesy($plots[2]) + imagesy($plots[3]) + $height[1] + $height[2] + 5;
  $image->img = imagecreate($image->width, $image->height);
  $image->background = imagecolorallocate($image->img, (float)$background_color[1], (float)$background_color[2], (float)$background_color[3]);
  $image->text = imagecolorallocate($image->img, (float)$text_color[1], (float)$text_color[2], (float)$text_color[3]);
  $image->margin = imagecolorallocate($image->img, (float)$margin_color[1], (float)$margin_color[2], (float)$margin_color[3]);
  imagefilledrectangle($image->img, 0, 0, $image->width, $image->height, $image->background);
  
  //
  // Make the title
  //
  imagefttext($image->img, 10, 0, ($image->width - $width[1]) / 2, 10, $image->text, $font_file, $rug_plot_title, array());

  //
  // Copy the individual plots into position
  //
  imagecopy($image->img, $plots[1], 0,                  $height[1],                                       0, 0, imagesx($plots[1]), imagesy($plots[1]));
  imagecopy($image->img, $plots[2], 0,                  imagesy($plots[1])+$height[1],                    0, 0, imagesx($plots[2]), imagesy($plots[2]));
  imagecopy($image->img, $plots[3], 0,                  imagesy($plots[1])+imagesy($plots[2])+$height[1], 0, 0, imagesx($plots[3]), imagesy($plots[3]));
  imagecopy($image->img, $plots[4], imagesx($plots[2]), $height[1],                                       0, 0, imagesx($plots[4]), imagesy($plots[4]));
  imagecopy($image->img, $plots[5], imagesx($plots[2]), imagesy($plots[4])+$height[1],                    0, 0, imagesx($plots[5]), imagesy($plots[5]));
  imagecopy($image->img, $plots[6], imagesx($plots[2]), imagesy($plots[4])+imagesy($plots[5])+$height[1], 0, 0, imagesx($plots[6]), imagesy($plots[6]));
  
  ////
  //// Stamp the run-time on the image. Debugging, of course.
  ////
  //$timer_2 = explode(' ', microtime());
  //$time = round($timer_2[0] + $timer_2[1] - $timer_1[0] - $timer_1[1], 5);
  //$rug_plot_footer .= " (created in $time seconds)";

  //
  // Make the footer
  //
  imagefttext($image->img, 8, 0, ($image->width - $width[2]) / 2, $image->height-4, $image->margin, $font_file, $rug_plot_footer, array());

  //
  // Directly display the result
  //
  //header("Content-type: image/png");
  //imagepng($image->img);

  //
  // Save the result to file
  //
  imagepng($image->img, $output_file);

  //
  // Clean up after ourselves
  //
  imagedestroy($image->img);
  imagedestroy($plots[1]);
  imagedestroy($plots[2]);
  imagedestroy($plots[3]);
  imagedestroy($plots[4]);
  imagedestroy($plots[5]);
  imagedestroy($plots[6]);
}
?>
